import asyncio
import json
import websocket
from typing import Any, Callable, Dict
from myko import MEvent, MEventType, MCommand, WSMEvent, WSMCommand

from .models import Target, Action, Emitter, Pulse, Instance, Machine, Alert, InstanceStatusEnum, TargetStatus, TargetStatusEnum, AlertEntityType, AlertLevel
from .utils import generate_hash, get_current_timestamp, make_instance_id

from .proxies.instance import InstanceProxy, InstanceArgs

from threading import Thread

class RshipExecClient:
    def __init__(self, rship_host: str, rship_port: int, on_exec_connected: Callable = None):
        self.rship_host = rship_host
        self.rship_port = rship_port
        self.is_connected = False
        self.client_id = None
        self.websocket = None
        self.targets: Dict[str, Target] = {}
        self.target_statuses: Dict[str, TargetStatus] = {}
        self.actions: Dict[str, Action] = {}
        self.emitters: Dict[str, Emitter] = {}
        self.instances: Dict[str, Instance] = {}
        self.machines: Dict[str, Machine] = {}
        self.alerts: Dict[str, Alert] = {}
        self.handlers: Dict[str, Callable[[Action, Any], None]] = {}

        self.instance_proxies: Dict[str, InstanceProxy] = {}
        self.on_exec_connected = on_exec_connected

    ##############################
    # Websockets
    ##############################

    def on_open(self, ws):
        print("Connected to Rship server successfully.")
        self.is_connected = True

    def on_message(self, ws, message):
        print(f"Received message: {message}")
        self.handle_message(message)

    def on_close(self, ws, status):
        print("Connection to Rship server closed.")
        self.is_connected = False

    def connect(self):
        uri = f"ws://{self.rship_host}:{self.rship_port}/myko"
        print(f"Attempting to connect to {uri}")
        try:
            self.websocket = websocket.WebSocketApp(
                uri, 
                on_message=lambda ws, message: self.on_message(ws, message), 
                on_open=lambda ws: self.on_open(ws),
                on_close=lambda ws, close_status_code: self.on_close(ws, close_status_code),
                on_error=lambda ws, error: print(f"Error: {error}")
            )

            self.ws_thread = Thread(target=self.websocket.run_forever)
            self.ws_thread.setDaemon(True)
            self.ws_thread.start()

        except Exception as e:
            print(f"Failed to connect: {e}")

    def disconnect(self):
        if self.websocket:
            self.websocket.close()
            self.websocket = None
            self.ws_thread = None
            print("Disconnected from Rship server")
            self.is_connected = False

    def send_event(self, event: MEvent):
        try: 
            self.websocket.send(json.dumps(WSMEvent(event).__dict__))
        except TypeError:
            print("Failed to send event: TypeError")

    def send_command(self, command: MCommand):
        try:
            self.websocket.send(json.dumps(WSMCommand(command).__dict__))
        except ConnectionError:
            print("Failed to send command: ConnectionError")

    def handle_message(self, message: str):
        data = json.loads(message)
        if data["event"] == "ws:m:command":
            command_data = data["data"]
            command_id = command_data["commandId"]
            if command_id == "client:setId":
                self.client_id = command_data["command"]["clientId"]
                print(f"Received client ID: {self.client_id}")
                if self.on_exec_connected:
                    self.on_exec_connected()
                # for instance_proxy in self.instance_proxies.values():
                #     instance_proxy.save(InstanceStatusEnum.Available)
            elif command_id == "target:action:exec":
                action_id = command_data["command"]["action"]["id"]
                if action_id in self.actions:
                    action = self.actions[action_id]
                    handler = self.handlers.get(action_id)
                    if handler:
                        handler(action, command_data["command"]["data"])

    ##############################
    # Models
    ##############################

    def set_data(self, item: Any, item_type: str):
        event = MEvent(item=item, item_type=item_type, change_type=MEventType.SET,
                       created_at=get_current_timestamp(), tx=generate_hash())
        self.send_event(event)


    def delete_data(self, item: Any):
        event = MEvent(item=item, item_type=type(item).__name__, change_type=MEventType.DEL,
                       created_at=get_current_timestamp(), tx=generate_hash())
        self.send_event(event)


    def add_instance(self, args: InstanceArgs):

        instance_proxy = InstanceProxy(args=args, 
                                       client=self)

        instance_proxy.save(InstanceStatusEnum.Available)
        self.instance_proxies[instance_proxy.id()] = instance_proxy   
        return instance_proxy


    def pulse_emitter(self, service_short_id: str, target_short_id: str, emitter_short_id: str, data: Any):
        full_emitter_id = f"{service_short_id}:{target_short_id}:{emitter_short_id}"
        pulse = Pulse(name="", emitter_id=full_emitter_id, data=data)
        self.set_data(pulse, 'Pulse')



from enum import Enum
from typing import List, Optional, Any, Callable
from myko import MItem, MCommand, Schema
from .utils import generate_hash

class Target(MItem):
    def __init__(self, id: str, name: str,
                 sub_targets: List[str], parent_targets: List[str], service_id: str,
                 bg_color: str, fg_color: str, last_updated: str, category: str, root_level: bool):
        super().__init__(id, name)

        self.category = category
        self.bgColor = bg_color
        self.fgColor = fg_color
        self.parentTargets = parent_targets
        self.subTargets = sub_targets
        self.serviceId = service_id
        self.lastUpdated = last_updated
        self.rootLevel = root_level
        # self.hash = str(uuid())


class TargetStatusEnum(Enum):
    Online = 'online'
    Offline = 'offline'


class TargetStatus(MItem):
    def __init__(self, id: str, name: str, target_id: str, status: TargetStatusEnum, last_updated: str, instance_id: str):
        super().__init__(id, name)

        self.targetId = target_id
        self.status = status.value
        self.lastUpdated = last_updated
        self.instanceId = instance_id
        # self.hash = generate_hash()


class Action(MItem):
    def __init__(self, id: str, name: str, schema: Optional[Schema], target_id: str, service_id: str):
        super().__init__(id, name)

        self.schema = schema
        self.targetId = target_id
        self.serviceId = service_id
        # self.hash = generate_hash()


class Emitter(MItem):
    def __init__(self, id: str, name: str, schema: Optional[Schema], target_id: str, service_id: str):
        super().__init__(id, name)

        self.schema = schema
        self.targetId = target_id
        self.serviceId = service_id
        # self.hash = generate_hash()


class Pulse(MItem):
    def __init__(self, name: str, emitter_id: str, data: Any):
        super().__init__(emitter_id, name)

        self.emitterId = emitter_id
        self.data = data
        # self.hash = generate_hash()


class InstanceStatusEnum(Enum):
    Starting = 'Starting'
    Available = 'Available'
    Stopping = 'Stopping'
    Unavailable = 'Unavailable'
    Error = 'Error'


class Instance(MItem):
    def __init__(self, id: str, name: str, service_id: str, client_id: str, cluster_id: str, service_type_code: str,
                 status: InstanceStatusEnum, machine_id: str, color: str, message: str):
        super().__init__(id, name)

        self.serviceId = service_id
        self.serviceTypeCode = service_type_code
        self.clientId = client_id
        self.clusterId = cluster_id
        self.machineId = machine_id
        self.color = color
        # self.hash = generate_hash()
        
        self.message = message
        self.status = status.value


class Machine(MItem):
    def __init__(self, name: str):
        super().__init__(name, name)

        self.name = name
        # self.hash = generate_hash()

class ExecTargetAction(MCommand):
    def __init__(self, tx: str, createdAt: str, action: Action, data: any):
        super().__init__(tx, createdAt)

        self.action = action
        self.data = data
        self.hash = generate_hash()


class AlertLevel(Enum):
    INFO = 'info'
    WARN = 'warn'
    ERROR = 'error'


class AlertEntityType(Enum):
    TARGET = 'Target'
    ACTION = 'Action'
    PAYLOAD = 'Payload'
    INSTANCE = 'Instance'


class Alert(MItem):
    def __init__(self, entity_id: str, entity_type: AlertEntityType, instance_id: str,
                 level: AlertLevel, message: str, code: str):
        super().__init__(entity_id + instance_id, f"{entity_id}:{code}")
        self.entityId = entity_id
        self.entityType = entity_type.value
        self.instanceId = instance_id
        self.level = level.value
        self.message = message
        self.code = code
        # self.hash = generate_hash()


from .client import RshipExecClient
from .proxies.instance import InstanceProxy, InstanceArgs
from .proxies.target import TargetProxy, TargetArgs
from .proxies.emitter import EmitterProxy, EmitterArgs




import uuid
from datetime import datetime, timezone
from typing import List

def generate_hash() -> str:
    return str(uuid.uuid4())

def get_current_timestamp() -> str:
    return datetime.now(timezone.utc).isoformat()

def make_instance_id(machine_id: str, service_id: str) -> str:
    return f"{machine_id}:{service_id}"

# def flatten_target_list(targets: List[TargetProps]) -> List[TargetProps]:
#     flattened = []
#     for target in targets:
#         flattened.append(target)
#         flattened.extend(flatten_target_list(target.subtargets))
#     return flattened